package us.blanshard.stubout;

import org.objectweb.asm.ClassAdapter;
import org.objectweb.asm.ClassVisitor;
import org.objectweb.asm.Label;
import org.objectweb.asm.MethodVisitor;
import org.objectweb.asm.Opcodes;
import org.objectweb.asm.Type;
import org.objectweb.asm.commons.AdviceAdapter;
import org.objectweb.asm.commons.Method;


/**
 * An ASM adapter that adds a check to every method allowing it to be replaced by a call
 * to a different method in a different class.
 * 
 * @author Luke Blanshard
 */
class MethodReplacingClassAdapter extends ClassAdapter
{
    private static final Type   STUBBER          = Type.getType(Stubber.class);
    private static final Method IS_REPLACED      = Method.getMethod("boolean isMethodReplaced(String, String)"); 
    private static final Method CALL_REPLACEMENT = Method.getMethod("Object callReplacement(String, String, Object, Object[])"); 

    private String className;

    public MethodReplacingClassAdapter(ClassVisitor cv) {
        super(cv);
    }

    public void visit( int version, int access, String name, String signature, String superName, String[] interfaces ) {
        className = name;
        super.visit(version, access, name, signature, superName, interfaces);
    }

    public MethodVisitor visitMethod( int access, String name, String desc, String signature, String[] exceptions ) {
        MethodVisitor mv = super.visitMethod(access, name, desc, signature, exceptions);
        if ( name.startsWith("<") ) // ie if it's a constructor or static initializer, don't allow replacements
            return mv;
        return new MethodReplacer( mv, access, name, desc );
    }
    
    private class MethodReplacer extends AdviceAdapter {
        private final boolean isStatic;
        private final Type returnType;
        private final String methodName;

        public MethodReplacer(MethodVisitor mv, int access, String name, String desc) {
            super(mv, access, name, desc);
            this.isStatic   = (access & Opcodes.ACC_STATIC) != 0;
            this.returnType = Type.getReturnType(desc);
            this.methodName = name; 
        }

        protected void onMethodEnter() {
            // Add calls to the Stubber methods isMethodReplaced and
            // callReplacement before the method body.
            String fullDesc = methodName + methodDesc;
            push(className);
            push(fullDesc);
            invokeStatic( STUBBER, IS_REPLACED );
            Label label = newLabel();
            ifZCmp(EQ, label); // If not replaced, jump to real method start

            push(className);
            push(fullDesc);
            if ( isStatic )
                push((String)null);
            else
                loadThis();
            loadArgArray();
            invokeStatic( STUBBER, CALL_REPLACEMENT );
            if ( returnType == Type.VOID_TYPE )
                pop();
            else
                unbox(returnType);
            returnValue();
            
            mark( label ); // Jump to here if not replacing
        }

        protected void onMethodExit( int opcode ) {
            // Nothing to do on exit
        }

        public void visitLocalVariable( String name, String desc, String signature, Label start, Label end, int index ) {
            // This works around odd behavior of EMMA-instrumented classes.
            try {
                if ( end.getOffset() > start.getOffset() )
                    super.visitLocalVariable(name, desc, signature, start, end, index);
            } catch (IllegalStateException e) {
                // This happens if one or the other labels has not been resolved,
                // meaning it's bogus.
            }
        }
    }
}
