package us.blanshard.stubout;

import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.objectweb.asm.ClassAdapter;
import org.objectweb.asm.ClassVisitor;
import org.objectweb.asm.FieldVisitor;
import org.objectweb.asm.MethodAdapter;
import org.objectweb.asm.MethodVisitor;
import org.objectweb.asm.Opcodes;

/**
 * An ASM adapter that renames the class it visits. You provide the old and new
 * names, and it substitutes accordingly as it goes. The names you give are
 * those of the outermost classes, and it will rename nested classes the same
 * way.
 * 
 * @author Luke Blanshard
 */
class RenamingClassAdapter extends ClassAdapter implements Opcodes
{
    private final Pattern oldNamePattern;
    private final String newName;

    /**
     * Creates a new adapter that decorates the given other visitor, and
     * substitutes the given new name for the given old name. Both of these
     * names are in "internal" form, meaning package names are separated by
     * forward slash characters (/) instead of dots. The visited class may match
     * the old name exactly, or it may be a nested class.
     * 
     * @param cv
     *            the class visitor to decorate with renaming capability
     * @param oldName
     *            the internal name of the outermost class being renamed
     * @param newName
     *            the internal name of the replacement name
     */
    public RenamingClassAdapter(ClassVisitor cv, String oldName, String newName) {
        super(cv);
        this.newName = quoteDollars(newName);
        this.oldNamePattern = Pattern.compile(quoteDollars(oldName));
    }

    public void visit( int version, int access, String name, String signature, String superName, String[] interfaces ) {
        super.visit(version, ACC_PUBLIC, fix(name), signature, superName, interfaces);
    }

    public MethodVisitor visitMethod( int access, String name, String desc, String signature, String[] exceptions ) {
        MethodVisitor mv = super.visitMethod(access, name, fix(desc), fix(signature), exceptions);
        if ( mv != null && (access & ACC_ABSTRACT) == 0 ) {
            mv = new MethodRenamer(mv);
        }
        return mv;
    }
    
    public FieldVisitor visitField( int access, String name, String desc, String signature, Object value ) {
        return super.visitField(access, name, fix(desc), fix(signature), value);
    }
    
    private static String quoteDollars( String s ) {
        if ( s.indexOf('$') >= 0 )
            s = s.replaceAll("\\$", "\\\\\\$"); // ie, replace $ with \$
        return s;
    }

    private String fix( String s ) {
        if ( s != null )
            s = oldNamePattern.matcher(s).replaceAll(newName);
        return s;
    }

    private class MethodRenamer extends MethodAdapter
    {
        public MethodRenamer(MethodVisitor mv) {
            super(mv);
        }

        public void visitTypeInsn( int i, String s ) {
            mv.visitTypeInsn(i, fix(s));
        }

        public void visitFieldInsn( int opcode, String owner, String name, String desc ) {
            mv.visitFieldInsn(opcode, fix(owner), name, fix(desc));
        }

        public void visitMethodInsn( int opcode, String owner, String name, String desc ) {
            mv.visitMethodInsn(opcode, fix(owner), name, fix(desc));
        }
    }
}
