package us.blanshard.stubout;

import java.io.IOException;
import java.io.InputStream;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.Map;

import org.objectweb.asm.ClassReader;
import org.objectweb.asm.ClassVisitor;
import org.objectweb.asm.ClassWriter;
import org.objectweb.asm.Type;

/**
 * Provides static methods that allow you to stub out entire classes or just
 * certain methods, from your JUnit test cases.
 * 
 * <p> StubOut test classes have the notion of a "before reloading" method (see {@link
 * StubbingTestCase#beforeReloading} for JUnit 3 tests or the {@link
 * StubbingRunner.BeforeReloading} annotation for JUnit 4 tests).  These methods are where
 * you must establish class replacements and stubbing exclusions by calling static methods
 * on Stubber: see the <code>replaceClass</code> methods, and <code>exclude</code> and
 * <code>excludePackage</code>.
 *
 * <p> To replace methods within a class, call <code>replaceMethods</code> from within a
 * unit test method.  
 * 
 * <p> The idea to stub out just certain methods came from Tobias Hill of Citerus AB (a
 * Swedish consultancy) -- see his description of <a
 * href="http://www.citerus.se/kunskap/pnehm/pnehmartiklar/5.ec944110677af1e83800018576.html">
 * JUnitAssist</a>.
 * 
 * @author Luke Blanshard
 */
public class Stubber
{
    /**
     * Called from a unit test's "before reloading" method to mark the given
     * original class as needing to be replaced by the given replacement. Note
     * that this method <b>cannot</b> be called from within a test method or
     * the setUp or tearDown methods.
     * 
     * @param original
     *            a class that must be replaced during the unit test
     * @param replacement
     *            the stub class that will replace it
     */
    public static void replaceClass( Class original, Class replacement ) {
        replaceClass(original.getName(), replacement.getName());
    }

    /**
     * Called from a unit test's "before reloading" method to mark the given
     * original class as needing to be replaced by the given replacement. Note
     * that this method <b>cannot</b> be called from within a test method or
     * the setUp or tearDown methods.
     * 
     * @param originalName
     *            the fully qualified name of a class that must be replaced
     *            during the unit test
     * @param replacement
     *            the stub class that will replace it
     */
    public static void replaceClass( String originalName, Class replacement ) {
        replaceClass(originalName, replacement.getName());
    }

    /**
     * Called from a unit test's "before reloading" method to mark the given
     * original class as needing to be replaced by the given replacement. Note
     * that this method <b>cannot</b> be called from within a test method or
     * the setUp or tearDown methods.
     * 
     * @param originalName
     *            the fully qualified name of a class that must be replaced
     *            during the unit test
     * @param replacementName
     *            the fully qualified name of the stub class that will replace
     *            it
     */
    public static void replaceClass( String originalName, String replacementName ) {
        if ( originalName == null || replacementName == null )
            throw new IllegalArgumentException("Neither the original nor the replacement class name may be null");
        if ( originalName.indexOf('$') >= 0 )
            throw new IllegalArgumentException("You may not replace a nested class");
        ThreadState state = (ThreadState) states.get();
        if ( state.stubbedClasses == null )
            throw new IllegalStateException(
                    "The Stubber.replaceClass method must be called from the beforeReloading "
                            + "method, not from setUp or a test method");
        if ( ! state.exclusions.shouldDefine(originalName) )
            throw new IllegalArgumentException( "You may not replace class "+originalName );
        state.stubbedClasses.put(originalName, replacementName);
    }

    /**
     * Called from a unit test class to have any methods defined by both the
     * given stubbed class and the given stub object, call the stub's versions
     * of the methods in place of the stubbed class's ones. This method must be
     * called from the setUp or test method.
     * 
     * <p>
     * Only methods actually defined by the stub object's class are used.
     * Inherited methods are not.
     * 
     * @param stubbedClass
     *            the class whose methods to replace
     * @param stub
     *            the object whose methods to use in their stead
     */
    public static void replaceMethods( Class stubbedClass, Object stub ) {
        if ( stubbedClass == null || stub == null )
            throw new IllegalArgumentException("Neither the stubbed class nor the stub may be null");
        if ( ! (stubbedClass.getClassLoader() instanceof Loader) )
            throw new IllegalArgumentException( "You may not replace methods from class "+stubbedClass.getName() );
        String key = stubbedClass.getName().replace('.', '/');
        synchronized (replacedMethods) {
            Map methods = (Map) replacedMethods.get(key);
            if ( methods == null ) {
                methods = new HashMap();
                replacedMethods.put(key, methods);
            }
            Method[] ms = stub.getClass().getDeclaredMethods();
            for ( int i = 0; i < ms.length; ++i ) {
                Method m = ms[i];
                m.setAccessible(true);
                String d = m.getName() + Type.getMethodDescriptor(m);
                methods.put(d, new StubbedMethod(stub, m));
            }
        }
    }

    /**
     * Undoes the replacement of methods by all previous calls to replaceMethods
     * for the given stubbed class.
     * 
     * @param stubbedClass
     *            the class whose replaced methods should be restored
     */
    public static void restoreMethods( Class stubbedClass ) {
        replacedMethods.remove(stubbedClass.getName().replace('.', '/'));
    }

    /**
     * Undoes all method replacements.
     */
    public static void restoreAllMethods() {
        replacedMethods.clear();
    }

    /**
     * A {@linkplain #replaceMethods(Class, Object) replacement method} may call
     * this to get the original object on which the method was called.
     */
    public static Object getOriginalTarget() {
        ThreadState state = (ThreadState) states.get();
        if ( state.originalTargets.isEmpty() )
            throw new IllegalStateException(
                    "An original target is only available within a replacement method");
        return state.originalTargets.getLast();
    }

    /**
     * A rule for excluding classes from stubbing.
     * @see Stubber#exclude(Exclusion)
     */
    public static interface Exclusion {
        /**
         * Tells the stubber whether the given fully qualified class name should
         * be excluded from stubbing.
         * 
         * @param className
         *            the fully qualified class name in question
         * @return flag telling whether to exclude the class from stubbing
         */
        public boolean shouldExclude( String className );
    }
    
    /**
     * Called from a unit test's "before reloading" method to exclude certain
     * classes from being reloaded by the stubber's class loader. Note that this
     * method <b>cannot</b> be called from within a test method or the setUp or
     * tearDown methods.
     * 
     * <p>
     * Knowing precisely when and how to apply an exclusion is a bit of a black
     * art. You will occasionally see an exception that makes no sense, like an
     * access or class-cast exception. If you are running a Sun JVM it is
     * sometimes necessary to exclude the "sun" package to make these exceptions
     * go away.
     * 
     * @param exclusion
     *            a rule excluding classes from stubbing
     * @see #excludePackage
     */
    public static void exclude( Exclusion exclusion ) {
        if ( exclusion == null )
            throw new IllegalArgumentException("The exclusion may not be null");
        ThreadState state = (ThreadState) states.get();
        if ( state.exclusions == null )
            throw new IllegalStateException(
                    "The Stubber.exclude method must be called from the beforeReloading "
                            + "method, not from setUp or a test method");
        state.exclusions.add(exclusion);
        for ( Iterator i=state.stubbedClasses.keySet().iterator(); i.hasNext(); ) {
            String className = (String) i.next();
            if ( ! state.exclusions.shouldDefine(className) ) {
                state.exclusions.remove( exclusion );
                throw new IllegalArgumentException( "You may not exclude a class you are replacing: "+className );
            }
        }
    }
    
    /**
     * Called from a unit test's "before reloading" method to exclude all
     * classes within the given package, including subpackages. For example,
     * <code>Stubber.excludePackage("javax.swing")</code> would exclude all
     * Swing classes.
     * 
     * <p>
     * Note that this method <b>cannot</b> be called from within a test method
     * or the setUp or tearDown methods.
     * 
     * @param packageName
     *            the root package to exclude
     * @see #exclude(Exclusion)
     */
    public static void excludePackage( String packageName ) {
        final String prefix = packageName + '.';
        exclude( new PrefixExclusion(prefix) );
    }

    /**
     * Part of the
     * {@linkplain #replaceMethods(Class, Object) method replacement} mechanism;
     * tells whether a given method has been replaced.
     */
    public static boolean isMethodReplaced( String className, String descriptor ) {
        synchronized (replacedMethods) {
            Map methods = (Map) replacedMethods.get(className);
            if ( methods == null )
                return false;
            return methods.containsKey(descriptor);
        }
    }

    /**
     * Part of the
     * {@linkplain #replaceMethods(Class, Object) method replacement} mechanism;
     * calls the replacement method.
     */
    public static Object callReplacement( String className, String descriptor, Object target, Object[] args )
            throws Throwable 
    {
        ThreadState state = (ThreadState) states.get();
        StubbedMethod m;
        synchronized (replacedMethods) {
            Map methods = (Map) replacedMethods.get(className);
            m = (StubbedMethod) methods.get(descriptor);
        }
        state.originalTargets.addLast(target);
        try {
            return m.invoke(args);
        } finally {
            state.originalTargets.removeLast();
        }
    }

    /**
     * Loads the given test's class through a stubbing class loader,
     * instantiates it, and executes the test case on the new object. Normally
     * only called from a TestCase subclass that implements ReloadableTest, such
     * as {@link StubbingTestCase}.
     * 
     * @param testCase
     *            the test case to run
     * @throws RuntimeException
     *             if the class can't be reloaded or reinstantiated
     * @throws Throwable
     *             if the reloaded test throws anything
     */
    public static void reloadAndRun( ReloadableTest testCase ) throws Throwable {
        ClassLoader loader = setUpReplacements(testCase);
        ReloadableTest newTestCase;
        try {
            Class c = loader.loadClass(testCase.getClass().getName());
            if ( c.getClassLoader() != loader )
                throw new IllegalArgumentException("Test class " + c.getName() + " can't be stubbed");
            newTestCase = (ReloadableTest) c.newInstance();
        } catch (RuntimeException e) {
            throw e;
        } catch (Exception e) {
            throw new RuntimeException("Unable to reload test class " + testCase.getName(), e);
        }
        restoreAllMethods();
        ClassLoader prev = Thread.currentThread().getContextClassLoader();
        Thread.currentThread().setContextClassLoader(loader);
        try {
            newTestCase.runBareTest(testCase.getName());
        } finally {
            Thread.currentThread().setContextClassLoader(prev);
        }
    }

    /**
     * Sets up any class replacements for the given test case, and returns an
     * appropriate class loader to reload the class.
     */
    static ClassLoader setUpReplacements( ReloadableTest testCase ) {
        ThreadState state = (ThreadState) states.get();
        state.stubbedClasses = new HashMap();
        state.exclusions = new Exclusions();
        try {
            testCase.beforeReloading();
            if ( state.loader == null 
                    || ! state.loader.stubbedClasses.equals(state.stubbedClasses) 
                    || ! state.loader.exclusions.equals(state.exclusions) )
                state.loader = new Loader(state.stubbedClasses, state.exclusions);
            return state.loader;
        } finally {
            state.stubbedClasses = null;
            state.exclusions = null;
        }
    }


    /** Can't construct this class. */
    private Stubber() {
    }


    private static final class PrefixExclusion implements Exclusion {
        private final String prefix;

        private PrefixExclusion(String prefix) {
            this.prefix = prefix;
        }

        public boolean shouldExclude(String className) {
            return className.startsWith(prefix);
        }
        
        public boolean equals( Object o ) {
            if ( ! (o instanceof PrefixExclusion) )
                return false;
            PrefixExclusion that = (PrefixExclusion) o;
            return this.prefix.equals(that.prefix);
        }
        
        public int hashCode() {
            return prefix.hashCode();
        }
    }

    /**
     * Holds a set of exclusions.
     */
    private static class Exclusions {
        private static final String OUR_PACKAGE_PREFIX = getPackagePrefix(Stubber.class);
        private static final String ASM_PACKAGE_PREFIX = getPackagePrefix(ClassReader.class);
        private final Collection exclusions = new LinkedHashSet();
        
        public void add( Exclusion e ) {
            exclusions.add( e );
        }
        
        public void remove( Exclusion e ) {
            exclusions.remove( e );
        }
        
        /**
         * Tells whether the loader should load the class with the given name.
         * This is the inverse of the Exclusion interface, which is a little
         * confusing, but it makes Loader easier to follow.
         */
        public boolean shouldDefine(String name) {
            // We start with the permanent exclusions:
            if ( name.startsWith("java.") || name.startsWith("junit.") || name.startsWith("org.junit.")
                    || name.startsWith(ASM_PACKAGE_PREFIX) )
                return false;
            if ( name.startsWith(OUR_PACKAGE_PREFIX) && getPackagePrefix(name).equals(OUR_PACKAGE_PREFIX) )
                return false;
            // And then we check the added exclusions:
            for ( Iterator i=exclusions.iterator(); i.hasNext(); ) {
                Exclusion e = (Exclusion) i.next();
                if ( e.shouldExclude(name) )
                    return false;
            }
            return true;
        }
        
        public boolean equals( Object o ) {
            if ( ! (o instanceof Exclusions) )
                return false;
            Exclusions that = (Exclusions) o;
            return this.exclusions.equals(that.exclusions);
        }

        public int hashCode() {
            return exclusions.hashCode();
        }

        private static String getPackagePrefix( Class c ) {
            return getPackagePrefix( c.getName() );
        }

        private static String getPackagePrefix( String s ) {
            int dot = s.lastIndexOf('.');
            return s.substring(0, dot + 1);
        }
    }


    /**
     * The class loader that instruments classes so their methods can be
     * replaced at runtime, and optionally replaces entire classes. The Stubber
     * class maintains one of these loaders statically.
     */
    private static class Loader extends ClassLoader
    {
        private final Map/* <String, String> */stubbedClasses;
        private final Exclusions exclusions;

        public Loader(Map stubbedClasses, Exclusions exclusions) {
            super(Stubber.class.getClassLoader());
            this.stubbedClasses = stubbedClasses;
            this.exclusions = exclusions;
        }

        protected synchronized Class loadClass( String name, boolean resolve ) throws ClassNotFoundException {
            // First, check if the class has already been loaded
            Class c = findLoadedClass(name);
            if ( c == null ) {
                if ( exclusions.shouldDefine(name) )
                    c = findClass(name);
                else if ( getParent() != null )
                    c = getParent().loadClass(name);
                else
                    c = findSystemClass(name);
            }
            if ( resolve ) {
                resolveClass(c);
            }
            return c;
        }

        protected Class findClass( String name ) throws ClassNotFoundException {
            String root = name;
            int dollar = name.indexOf('$');
            if ( dollar > 0 )
                root = name.substring(0, dollar);
            String subst = (String) stubbedClasses.get(root);
            String substRoot = subst;
            if ( subst != null && dollar > 0 )
                subst = subst + name.substring(dollar);
            String path = (subst == null) ? name : subst;
            path = path.replace('.', '/') + ".class";
            byte[] b;
            try {
                InputStream in = getResourceAsStream(path);
                ClassReader cr = new ClassReader(in);
                ClassWriter cw = new ClassWriter(ClassWriter.COMPUTE_MAXS);
                ClassVisitor cv = cw;
                cv = new MethodReplacingClassAdapter(cv);
                if ( subst != null )
                    cv = new RenamingClassAdapter(cv, substRoot.replace('.', '/'), root.replace('.', '/'));
                cr.accept(cv, 0);
                b = cw.toByteArray();
            } catch (IOException e) {
                throw new ClassNotFoundException(name, e);
            }
            return defineClass(name, b, 0, b.length);
        }
    }

    /**
     * Encapsulates a replacement method and the object that defines it.
     */
    private static class StubbedMethod
    {
        private final Object stub;
        private final Method method;

        public StubbedMethod(Object stub, Method method) {
            this.stub = stub;
            this.method = method;
        }

        public Object invoke( Object[] args ) throws Throwable {
            try {
                return method.invoke(stub, args);
            } catch (InvocationTargetException e) {
                throw e.getCause();
            }
        }
    }

    /**
     * Holds per-thread state.
     */
    private static class ThreadState
    {
        public Map stubbedClasses;
        public Exclusions exclusions;
        public Loader loader;
        public final LinkedList originalTargets = new LinkedList();
    }

    private static final ThreadLocal/* <ThreadState> */states = new ThreadLocal() {
        protected Object initialValue() {
            return new ThreadState();
        }
    };


    private static final Map/* <String, Map<String, StubbedMethod>> */replacedMethods = Collections
            .synchronizedMap(new HashMap());
}
