package us.blanshard.stubout;

import java.lang.annotation.ElementType;
import java.lang.annotation.Inherited;
import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;
import java.lang.annotation.Target;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

import org.junit.BeforeClass;
import org.junit.internal.runners.TestClassRunner;
import org.junit.internal.runners.TestIntrospector;
import org.junit.runner.Description;
import org.junit.runner.Request;
import org.junit.runner.Runner;
import org.junit.runner.manipulation.Filter;
import org.junit.runner.manipulation.Filterable;
import org.junit.runner.manipulation.NoTestsRemainException;
import org.junit.runner.manipulation.Sortable;
import org.junit.runner.manipulation.Sorter;
import org.junit.runner.notification.Failure;
import org.junit.runner.notification.RunListener;
import org.junit.runner.notification.RunNotifier;

/**
 * The JUnit 4 test runner for StubOut. Test classes that use StubOut must be
 * annotated with <code>@RunWith(StubbingRunner.class)</code>.
 * 
 * <p>
 * StubbingRunner also declares its own nested {@link RunWith} annotation; you
 * may annotate your class with it too, to have the reloaded test class run with
 * the given (other) runner.
 * 
 * <p>

 * And StubbingRunner declares another annotation, {@link BeforeReloading}. Use this like
 * the {@link BeforeClass} JUnit 4 annotation to mark a <code>public static void</code>
 * method as containing calls to {@link Stubber#replaceClass}.
 * 
 * @author Luke Blanshard
 */
public class StubbingRunner extends Runner implements Filterable, Sortable 
{
    /**
     * Defines a runner for the reloaded test class. That is, the StubbingRunner
     * will use this annotation to decide which other runner to use after it has
     * worked its stubbing magic. If this is not specified, it will use the
     * default JUnit 4 runner.
     */
    @Retention(RetentionPolicy.RUNTIME)
    @Target(ElementType.TYPE)
    @Inherited
    public @interface RunWith {
        Class<? extends Runner> value();
    }

    /**
     * Marks a <code>public static void</code> method as needing to be run before the
     * stubber reloads the test class; one of these methods is where you may call {@link
     * Stubber#replaceClass(Class,Class)} or {@link Stubber#excludePackage(String)}. This
     * takes the place of the {@link StubbingTestCase#beforeReloading()} method in the
     * JUnit 3 test classes. The <code>BeforeReloading</code> methods of superclasses are
     * run before those of the current class.
     */
    @Retention(RetentionPolicy.RUNTIME)
    @Target(ElementType.METHOD)
    public @interface BeforeReloading {
    }

    public StubbingRunner(Class testClass) {
        this.testClass = testClass;
    }

    @Override
    public Description getDescription() {
        return getStubbingRunner(testClass).getDescription();
    }

    @Override
    public void run(final RunNotifier notifier) {
        // This method has a lot in common with Stubber.reloadAndRun.
        ClassLoader loader = Stubber.setUpReplacements(new ClassReplacer(
                notifier));
        Class reloadedTestClass;
        try {
            reloadedTestClass = loader.loadClass(testClass.getName());
            if (reloadedTestClass.getClassLoader() != loader)
                throw new IllegalArgumentException("Test class "
                        + testClass.getName() + " can't be stubbed");
        } catch (RuntimeException e) {
            notifier.fireTestFailure(new Failure(getDescription(), e));
            throw e;
        } catch (Exception e) {
            notifier.fireTestFailure(new Failure(getDescription(), e));
            throw new RuntimeException("Unable to reload test class "
                    + testClass.getName(), e);
        }

        // On each individual test that's run, clear out any method replacements
        // before the test starts.
        RunListener listener = new RunListener() {
            @Override
            public void testStarted(Description description) {
                Stubber.restoreAllMethods();
            }
        };
        ClassLoader prev = Thread.currentThread().getContextClassLoader();
        Thread.currentThread().setContextClassLoader(loader);
        notifier.addListener(listener);
        try {
            getStubbingRunner(reloadedTestClass).run(notifier);
        } finally {
            notifier.removeListener(listener);
            Thread.currentThread().setContextClassLoader(prev);
        }
    }

    public void filter(final Filter filter) {
        if (filter == null)
            throw new IllegalArgumentException("Null filter not allowed");
        if (this.filter == null)
            this.filter = filter;
        else
            this.filter = new Filter() {
                private final Filter a = StubbingRunner.this.filter;
                private final Filter b = filter;
                @Override
                public String describe() {
                    return a.describe() + " and " + b.describe();
                }

                @Override
                public boolean shouldRun(Description description) {
                    return a.shouldRun(description) && b.shouldRun(description);
                }
            };
    }

    public void sort(final Sorter sorter) {
        if (sorter == null)
            throw new IllegalArgumentException("Null sorter not allowed");
        if (this.sorter == null)
            this.sorter = sorter;
        else {
            Comparator<Description> comp = new Comparator<Description>() {
                // Note that the new sorter becomes primary, opposite of the
                // filter case.
                private final Sorter a = sorter;
                private final Sorter b = StubbingRunner.this.sorter;
                public int compare(Description o1, Description o2) {
                    int answer = a.compare(o1, o2);
                    if (answer == 0)
                        answer = b.compare(o1, o2);
                    return answer;
                }
            };
            this.sorter = new Sorter(comp);
        }
    }

    private Runner getStubbingRunner(Class<?> testClass) {
        RunWith annotation = testClass.getAnnotation(RunWith.class);
        Class<? extends Runner> runnerClass;
        if (annotation == null)
            runnerClass = TestClassRunner.class;
        else
            runnerClass = annotation.value();
        try {
            Constructor<? extends Runner> ctor = runnerClass
                    .getConstructor(Class.class);
            Runner answer = ctor.newInstance(testClass);
            if (filter != null)
                filter.apply(answer);
            if (sorter != null)
                sorter.apply(answer);
            return answer;
        } catch (Exception e) {
            return Request.errorReport(testClass, e).getRunner();
        }
    }

    private final class ClassReplacer implements ReloadableTest {
        private final RunNotifier notifier;

        private ClassReplacer(RunNotifier notifier) {
            this.notifier = notifier;
        }

        public String getName() {
            return null;
        }

        public void runBareTest(String name) throws Throwable {
        }

        public void beforeReloading() {
            List<Method> methods = new TestIntrospector(testClass)
                    .getTestMethods(BeforeReloading.class);
            Collections.reverse(methods); // This makes the superclass methods come first
            for (Method m : methods)
                try {
                    m.invoke(null);
                } catch (InvocationTargetException e) {
                    fail(e.getCause());
                } catch (Throwable t) {
                    fail(t);
                }
        }

        private void fail(Throwable t) {
            notifier.fireTestFailure(new Failure(getDescription(), t));
            throw new RuntimeException(
                    "Unable to replace classes for test class "
                            + testClass.getName(), t);
        }
    }

    /**
     * The class under test by this runner, before it's been reloaded by
     * Stubber's class loader.
     */
    private final Class testClass;

    private Filter filter;
    private Sorter sorter;
}
